/*-----------------------------------------------------------------*/
/*! 
  \file calcephinpopio.c 
  \brief perform I/O on INPOP/original JPL ephemeris data file 
         interpolate INPOP and JPL Ephemeris data.
         compute position and velocity from binary ephemeris file.

  \author  M. Gastineau 
           Astronomie et Systemes Dynamiques, IMCCE, CNRS, Observatoire de Paris. 

   Copyright, 2008, 2009, 2010, 2011, CNRS
   email of the author : gastineau@imcce.fr

  History:                                                                
*/
/*-----------------------------------------------------------------*/

/*-----------------------------------------------------------------*/
/* License  of this file :
 This file is "triple-licensed", you have to choose one  of the three licenses 
 below to apply on this file.
 
    CeCILL-C
    	The CeCILL-C license is close to the GNU LGPL.
    	( http://www.cecill.info/licences/Licence_CeCILL-C_V1-en.html )
   
 or CeCILL-B
        The CeCILL-B license is close to the BSD.
        (http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.txt)
  
 or CeCILL v2.0
      The CeCILL license is compatible with the GNU GPL.
      ( http://www.cecill.info/licences/Licence_CeCILL_V2-en.html )
 

This library is governed by the CeCILL-C, CeCILL-B or the CeCILL license under 
French law and abiding by the rules of distribution of free software.  
You can  use, modify and/ or redistribute the software under the terms 
of the CeCILL-C,CeCILL-B or CeCILL license as circulated by CEA, CNRS and INRIA  
at the following URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C,CeCILL-B or CeCILL license and that you accept its terms.
*/
/*-----------------------------------------------------------------*/

#include "calcephconfig.h"
#if HAVE_STDIO_H
#include <stdio.h>
#endif
#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if HAVE_STRING_H
#include <string.h>
#endif
#if HAVE_ERRNO_H
#include <errno.h>
#endif
#if HAVE_MATH_H
#include <math.h>
#endif
#if HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#include "calcephdebug.h"
#include "real.h"
#define __CALCEPH_WITHIN_CALCEPH 1
#include "calceph.h"
#include "calcephinternal.h"
#include "util.h"

/*==========================================================================*/
/* Fonctions locales                                                        */
/*==========================================================================*/
static int calceph_interpol_PV_planet(struct calcephbin_inpop *p_pbinfile, treal TimeJD0, treal Timediff,
                                      int Target, int *Unit, stateType * Planet);
static int calceph_inpop_unit_convert(struct calcephbin_inpop *pephbin, stateType * Planet,
                                      int Target, int InputUnit, int OutputUnit);

/*--------------------------------------------------------------------------*/
/*!                                                                       
     This function is used by the functions below to read an array of     
     Tchebeychev coefficients from a binary ephemeris data file.          
  @return return 1 if no error occurs. return 0 if an error occurs. 
  
  @param pephbin(inout) descriptor of the ephemerids
  @param Time (in) Desired record time.   
*/
/*--------------------------------------------------------------------------*/
int calceph_inpop_readcoeff(t_memarcoeff * pcoeftime_array, double Time)
{
    int n = fread(pcoeftime_array->Coeff_Array, sizeof(double), pcoeftime_array->ARRAY_SIZE, pcoeftime_array->file);

    if (n != pcoeftime_array->ARRAY_SIZE)
    {
        fatalerror("Can't read ephemeris file at time=%g. System error: '%s'\n", Time, strerror(errno));
        return 0;
    }
    /* swap the data if necessary */
    if (pcoeftime_array->swapbyteorder)
        swapdblarray(pcoeftime_array->Coeff_Array, pcoeftime_array->ARRAY_SIZE);

    pcoeftime_array->T_beg = pcoeftime_array->Coeff_Array[0];
    pcoeftime_array->T_end = pcoeftime_array->Coeff_Array[1];
    pcoeftime_array->T_span = pcoeftime_array->T_end - pcoeftime_array->T_beg;

    /* verifie s'il y a eu une erreur de fread ou de fseek */
    if (Time < pcoeftime_array->T_beg || Time > pcoeftime_array->T_end)
    {
        fatalerror("Read a bad block [%.16g , %.16g ] in the ephemeris file at the time %g\n",
                   pcoeftime_array->T_beg, pcoeftime_array->T_end, Time);
        return 0;
    }
    return 1;
}

/*--------------------------------------------------------------------------*/
/*!                                                                       
     Seek the record containing the time.
     Read an array of Tchebychev coefficients from a binary ephemeris data file.   
     
  @return return true if time is available 
  
  @param pephbin(inout) coefficient array + binary file
  @param Time (in) Desired record time.   
  
*/
/*--------------------------------------------------------------------------*/
int calceph_inpop_seekreadcoeff(t_memarcoeff * pephbin, double Time)
{
    double T_delta = 0.0;
    off_t Offset = 0;

  /*--------------------------------------------------------------------------*/
    /*  Find ephemeris data that record contains input time. Note that one, and */
    /*  only one, of the following conditional statements will be true (if both */
    /*  were false, this function would not have been called).                  */
  /*--------------------------------------------------------------------------*/

    if (Time < pephbin->T_beg)  /* Compute backwards location offset */
    {
        T_delta = pephbin->T_beg - Time;
        Offset = (int) -ceil(T_delta / pephbin->T_span);
    }

    if (Time > pephbin->T_end)  /* Compute forwards location offset */
    {
        T_delta = Time - pephbin->T_end;
        Offset = (int) ceil(T_delta / pephbin->T_span);
    }

  /*--------------------------------------------------------------------------*/
    /*  Retrieve ephemeris data from new record.                                */
  /*--------------------------------------------------------------------------*/
    pephbin->offfile += Offset * pephbin->ARRAY_SIZE * sizeof(double);
    if (fseeko(pephbin->file, pephbin->offfile, SEEK_SET) != 0)
    {
        fatalerror("Can't jump in the file at time=%g. System error: '%s'\n", Time, strerror(errno));
        return 0;
    }
#if DEBUG
    if (ftello(pephbin->file) != pephbin->offfile)
    {
        printf("error in the file location %lu %lu\n",
               (unsigned long) pephbin->offfile, (unsigned long) ftello(pephbin->file));
    }
    printf("set to the location %lu\n", (unsigned long) pephbin->offfile);
#endif
    return calceph_inpop_readcoeff(pephbin, Time);
}

/*--------------------------------------------------------------------------*/
/*!  
     Open the "filename" ephemeris file .
     Read the header data block
     Read the constant data block
     Read the first coefficient data bock.
     
  @return descriptor of the ephemeris and the first array of coefficients   
  
  @param fileName (in) A character string giving the name of an ephemeris data file.   
*/
/*--------------------------------------------------------------------------*/
int calceph_inpop_open(FILE * file, const char *fileName, struct calcephbin_inpop *res)
{
    int computerecordsize = 0;  /* = 1 => require to compute the record size */
    int j;
    int newdenum;

    res->coefftime_array.Coeff_Array = NULL;
    res->coefftime_array.file = file;
    res->timescale = etimescale_TDB;

  /*--------------------------------------------------------------------------*/
    /*  Intialize the asteroid information to empty .                          */
  /*--------------------------------------------------------------------------*/
    calceph_empty_asteroid(&res->asteroids);

  /*--------------------------------------------------------------------------*/
    /*  Read header & first coefficient array, then return status code.         */
  /*--------------------------------------------------------------------------*/
    if (fseeko(file, 0, SEEK_SET) != 0)
    {
        fatalerror("Can't jump to the beginning of the file '%s'. System error: '%s'\n", fileName, strerror(errno));
        calceph_inpop_close(res);
        return 0;
    }

    /*Read first three header records from ephemeris file */

    if (fread(&res->H1, sizeof(t_HeaderBlock), 1, file) != 1)
    {
        fatalerror("Can't read header block from the ephemeris file '%s'\nSystem error : '%s'\n",
                   fileName, strerror(errno));
        calceph_inpop_close(res);
        return 0;
    }
    res->coefftime_array.ncomp = 3;
    res->coefftime_array.swapbyteorder = false;
    res->isinAU = false;
    res->isinDay = true;
    res->haveTTmTDB = false;
    switch (res->H1.DENUM)
    {
        case EPHEMERIS_INPOP:
            res->coefftime_array.ARRAY_SIZE = res->H1.recordsize;
            res->coefftime_array.ncomp = 6;
            break;
        default:
            switch (swapint(res->H1.DENUM))
            {
                case EPHEMERIS_INPOP:
                    res->coefftime_array.swapbyteorder = true;
                    res->coefftime_array.ARRAY_SIZE = res->H1.recordsize;
                    if (res->coefftime_array.swapbyteorder)
                        res->coefftime_array.ARRAY_SIZE = swapint(res->coefftime_array.ARRAY_SIZE);
                    res->coefftime_array.ncomp = 6;
                    break;
                default:
                    res->coefftime_array.ARRAY_SIZE = 0;
                    newdenum = res->H1.DENUM;
                    if (((unsigned int) res->H1.DENUM) > (1U << 15))
                    {
                        newdenum = swapint(res->H1.DENUM);
                        res->coefftime_array.swapbyteorder = true;
                        computerecordsize = 1;
                    }
                    if (newdenum >= EPHEMERIS_DE200)
                    {
                        computerecordsize = 1;
                    }
                    else
                    {
                        fatalerror("Opened wrong file: '%s' for ephemeris: DENUM=%d.\n", fileName, newdenum);
                        calceph_inpop_close(res);
                        return 0;
                    }
                    break;
            }
            break;
    }

   /*--------------------------------------------------------------------------*/
    /* swap the header if necessary */
   /*--------------------------------------------------------------------------*/
    if (res->coefftime_array.swapbyteorder)
    {
        swapdblarray(res->H1.timeData, 3);
        res->H1.numConst = swapint(res->H1.numConst);
        res->H1.AU = swapdbl(res->H1.AU);
        res->H1.EMRAT = swapdbl(res->H1.EMRAT);
        swapintarray((int *) res->H1.coeffPtr, 36);
        res->H1.DENUM = swapint(res->H1.DENUM);
        swapintarray(res->H1.libratPtr, 3);
        swapintarray(res->H1.TTmTDBPtr, 3);
        res->H1.recordsize = swapint(res->H1.recordsize);
    }

   /*--------------------------------------------------------------------------*/
    /* compute record size if required */
   /*--------------------------------------------------------------------------*/
    if (computerecordsize == 1)
    {
        int failure = 0;

        if (res->coefftime_array.ncomp <= 0)
            failure = 1;
        for (j = 0; j < 12; j++)
        {
            int ncompcur = (j == 11) ? 2 : res->coefftime_array.ncomp;

            if (res->H1.coeffPtr[j][1] < 0 || res->H1.coeffPtr[j][2] < 0)
                failure = 1;
            res->coefftime_array.ARRAY_SIZE += res->H1.coeffPtr[j][1] * res->H1.coeffPtr[j][2] * ncompcur;
        }
        if (res->H1.libratPtr[1] < 0 || res->H1.libratPtr[2] < 0)
            failure = 1;
        res->coefftime_array.ARRAY_SIZE += res->H1.libratPtr[1] * res->H1.libratPtr[2] * res->coefftime_array.ncomp;
        res->coefftime_array.ARRAY_SIZE += 2;
        res->H1.recordsize = res->coefftime_array.ARRAY_SIZE;

        if (failure || res->coefftime_array.ARRAY_SIZE < 0)
        {
            fatalerror("Opened wrong file: '%s' for ephemeris: DENUM=%d.\n", fileName, res->H1.DENUM);
            calceph_inpop_close(res);
            return 0;
        }
        //printf("recordsize = %d\n", res->ARRAY_SIZE);
    }

   /*--------------------------------------------------------------------------*/
    /* read constants */
   /*--------------------------------------------------------------------------*/
    res->coefftime_array.Coeff_Array = (double *) malloc(sizeof(double) * res->coefftime_array.ARRAY_SIZE);
    if (res->coefftime_array.Coeff_Array == NULL)
    {
        fatalerror("Can't allocate memory for %d reals\nSystem error : '%s'\n",
                   res->coefftime_array.ARRAY_SIZE, strerror(errno));
        calceph_inpop_close(res);
        return 0;

    }
    /*lecture des constantes */
    /*se positionne au debut du deuxieme record */
    if (fseeko(file, sizeof(double) * res->coefftime_array.ARRAY_SIZE, SEEK_SET) != 0)
    {
        fatalerror("Can't jump to constant values in the ephemeris file '%s'\nSystem error : '%s'\n",
                   fileName, strerror(errno));
        calceph_inpop_close(res);
        return 0;
    }
    if (fread(res->constVal, sizeof(double), 400, file) != 400)
    {
        fatalerror("Can't read constant values from the ephemeris file '%s'\nSystem error : '%s'\n",
                   fileName, strerror(errno));
        calceph_inpop_close(res);
        return 0;
    }
    /* swap the data if necessary */
    if (res->coefftime_array.swapbyteorder)
        swapdblarray(res->constVal, 400);

    /* ajuste le nombre de composantes pour inpop */
    if (res->H1.DENUM == EPHEMERIS_INPOP)
    {
        double dformat = 0.E0;
        double dunite;
        double dtimescale;

        if (calceph_inpop_getconstant(res, "FORMAT", &dformat))
        {
            if (((int) (fmod(dformat, 10))) == 1)
            {
                res->coefftime_array.ncomp = 3;
            };
            if (((int) (fmod(rint(dformat / 10), 10))) == 1)
            {
                res->haveTTmTDB = true;
            };
     /*--------------------------------------------------------------------------*/
            /*  initialize the asteroids data */
     /*--------------------------------------------------------------------------*/
            if (((int) (fmod(rint(dformat / 100), 100))) == 1)
            {
                double SS[3];

                for (j = 0; j <= 2; j++)
                    SS[j] = res->H1.timeData[j];
                if (calceph_init_asteroid
                    (&res->asteroids, res->coefftime_array.file, res->coefftime_array.swapbyteorder, SS,
                     res->H1.recordsize, res->coefftime_array.ncomp) == 0)
                {
                    calceph_inpop_close(res);
                    return 0;
                }
            }
        }

        if (calceph_inpop_getconstant(res, "UNITE", &dunite))
        {
            int iunite = (int) dunite;

            res->isinAU = (iunite == 0);
            res->isinDay = (iunite == 0 || iunite == 1);
        }
        else
        {
            fatalerror("Can't find constant 'UNITE' in ephemeris file. Wrong version of ephemeris file\n");
            calceph_inpop_close(res);
            return 0;
        }

    /*--------------------------------------------------------------------------*/
        /*  initialize the time scale */
    /*--------------------------------------------------------------------------*/
        if (calceph_inpop_getconstant(res, "TIMESC", &dtimescale))
        {
            res->timescale = (t_etimescale) dtimescale;
        }

    }

  /*--------------------------------------------------------------------------*/
    /*  read the third record : first records for the planets */
  /*--------------------------------------------------------------------------*/
    res->coefftime_array.offfile = 2 * sizeof(double) * res->coefftime_array.ARRAY_SIZE;
    if (fseeko(file, res->coefftime_array.offfile, SEEK_SET))
    {
        fatalerror("Can't jump to first data record in the ephemeris file '%s'\nSystem error : '%s'\n",
                   fileName, strerror(errno));
        calceph_inpop_close(res);
        return 0;
    }
    if (!calceph_inpop_readcoeff(&res->coefftime_array, res->H1.timeData[0]))
    {
        calceph_inpop_close(res);
        return 0;
    }

#if DEBUG
    {
        printf("\n  calceph_open : \n");
        printf("\n      ARRAY_SIZE = %4d", res->coefftime_array.ARRAY_SIZE);
        printf("\n      T_Beg      = %7.3f", res->coefftime_array.T_beg);
        printf("\n      T_End      = %7.3f", res->coefftime_array.T_end);
        printf("\n      T_Span     = %7.3f", res->coefftime_array.T_span);
        printf("\n      Number of component     = %d", res->coefftime_array.ncomp);
        printf("\n      isinAU     = %d", res->isinAU);
        printf("\n      timescale     = %d", res->isinAU);
        printf("\n      TT-TDB     = %d", res->timescale);
        printf("\n      record size     = %d\n\n", res->H1.recordsize);
    }
#endif /*DEBUG*/
        return 1;
}

/*--------------------------------------------------------------------------*/
/*! Close the  ephemeris file and destroy the ephemeris descriptor
 @param eph (inout) ephemeris descriptor to be destroyed
*/
/*--------------------------------------------------------------------------*/
void calceph_inpop_close(struct calcephbin_inpop *eph)
{
    if (eph->coefftime_array.file != NULL)
        fclose(eph->coefftime_array.file);
    if (eph->coefftime_array.Coeff_Array != NULL)
        free(eph->coefftime_array.Coeff_Array);
    eph->coefftime_array.file = NULL;
    eph->coefftime_array.Coeff_Array = NULL;
    calceph_free_asteroid(&eph->asteroids);
}

/*--------------------------------------------------------------------------*/
/*!                                                      
     This function reads the data block corresponding to the specified Time
     if this block isn't already in memory.
     
     This function works only on all targets.
     for asteroids, the target value is "asteroid number"+CALCEPH_ASTEROID
     
     It computes position and velocity vectors for a selected  
     planetary body from Chebyshev coefficients.          

                                                                          
  @param pephbin (inout) descripteur de l'ephemeride                            
  @param TimeJD0+timediff  (in) Time for which position is desired (Julian Date)
  @param Target   (in) Solar system body for which position is desired.         
  @param Center   (in) Solar system body from which position is desired.         
  @param Position (in) Pointer to external array to receive the position.       
  @param Unit  (in) output unit of the statetype
  @param Planet (out) position and veloticity                                                                        
         the units are expressed in Unit units.
*/
/*--------------------------------------------------------------------------*/
int calceph_inpop_interpol_PV(struct calcephbin_inpop *pephbin, treal TimeJD0, treal Timediff,
                              int Target, int Center, int Unit, stateType * Planet)
{
    treal Time = TimeJD0 + Timediff;
    int res = 0;
    int ephunit;

    /* check if time is in the ephemeris */
    if (Time < pephbin->H1.timeData[0] || Time > pephbin->H1.timeData[1])
    {
        fatalerror("time %g isn't in the interval [%.16g , %.16g] \n",
                   Time, pephbin->H1.timeData[0], pephbin->H1.timeData[1]);
        return 0;
    }

    if (Target >= CALCEPH_ASTEROID)
    {                           /* add 1 to target because Target is decreased by calceph_compute */
        res =
            calceph_interpol_PV_asteroid(&pephbin->asteroids, TimeJD0, Timediff, Target + 1, &ephunit, pephbin->isinAU,
                                         Planet);
    }
    else
    {
        res = calceph_interpol_PV_planet(pephbin, TimeJD0, Timediff, Target, &ephunit, Planet);
    }
    if (res)
        res = calceph_inpop_unit_convert(pephbin, Planet, Target, ephunit, Unit);
    return res;
}

/*--------------------------------------------------------------------------*/
/*!                                                      
     This function reads the data block corresponding to the specified Time
     if this block isn't already in memory.
     
     This function works only on planets/libration/moon/TT-TDB.
     
     It computes position and velocity vectors for a selected  
     planetary body from Chebyshev coefficients.          

                                                                          
  @param p_pbinfile (inout) descripteur de l'ephemeride                            
  @param TimeJD0+timediff  (in) Time for which position is desired (Julian Date)
  @param Target   (in) Solar system body for which position is desired.         
  @param Position (in) Pointer to external array to receive the position.       
  @param ephunit  (out) output unit of the statetype
  @param Planet (out) position and veloticity                                                                        
         the units are expressed in ephunit units.
*/
/*--------------------------------------------------------------------------*/
static int calceph_interpol_PV_planet(struct calcephbin_inpop *p_pbinfile, treal TimeJD0, treal Timediff,
                                      int Target, int *ephunit, stateType * Planet)
{
    treal Time = TimeJD0 + Timediff;
    long int C, G, N;

    if (Target == NUTATIONS)
    {
        fatalerror("\n This function does not compute nutations.\n");
        return 0;
    }

  /*--------------------------------------------------------------------------*/
    /* Determine if a new record needs to be input.                             */
  /*--------------------------------------------------------------------------*/

    if (Time < p_pbinfile->coefftime_array.T_beg || Time > p_pbinfile->coefftime_array.T_end)
    {
        if (!calceph_inpop_seekreadcoeff(&p_pbinfile->coefftime_array, Time))
            return 0;
    }

  /*--------------------------------------------------------------------------*/
    /* Read the coefficients from the binary record.                            */
  /*--------------------------------------------------------------------------*/

    if (Target == LIBRATIONS)
    {                           /*Librations */
        C = p_pbinfile->H1.libratPtr[0] - 1;
        N = p_pbinfile->H1.libratPtr[1];
        G = p_pbinfile->H1.libratPtr[2];
        *ephunit = CALCEPH_UNIT_RAD + (p_pbinfile->isinDay ? CALCEPH_UNIT_DAY : CALCEPH_UNIT_SEC);
    }
    else if (Target == TTMTDB)
    {                           /*TT-TDB */
#if DEBUG
        printf("target=TTMTDB\n");
#endif
        if (!p_pbinfile->haveTTmTDB)
        {
            fatalerror(" The file doesn't have TT-TDB data.\n");
            return 0;
        }
        if (p_pbinfile->timescale != etimescale_TDB)
        {
            fatalerror(" The file is not expressed in the TDB time scale.\n");
            return 0;
        }
        C = p_pbinfile->H1.TTmTDBPtr[0] - 1;
        N = p_pbinfile->H1.TTmTDBPtr[1];
        G = p_pbinfile->H1.TTmTDBPtr[2];
        *ephunit = CALCEPH_UNIT_SEC;
    }
    else if (Target == TCGMTCB)
    {                           /*TCG-TCB */
#if DEBUG
        printf("target=TCGMTCB p_pbinfile->timescale=%d\n", (int) p_pbinfile->timescale);
#endif
        if (!p_pbinfile->haveTTmTDB)
        {
            fatalerror(" The file doesn't have TCG-TCB data.\n");
            return 0;
        }
        if (p_pbinfile->timescale != etimescale_TCB)
        {
            fatalerror(" The file is not expressed in the TCB time scale.\n");
            return 0;
        }
        C = p_pbinfile->H1.TTmTDBPtr[0] - 1;
        N = p_pbinfile->H1.TTmTDBPtr[1];
        G = p_pbinfile->H1.TTmTDBPtr[2];
        *ephunit = CALCEPH_UNIT_SEC;
    }
    else
    {                           /* Planets */
        C = p_pbinfile->H1.coeffPtr[Target][0] - 1;
        N = p_pbinfile->H1.coeffPtr[Target][1];
        G = p_pbinfile->H1.coeffPtr[Target][2];
        *ephunit = (p_pbinfile->isinAU ? CALCEPH_UNIT_AU : CALCEPH_UNIT_KM)
            + (p_pbinfile->isinDay ? CALCEPH_UNIT_DAY : CALCEPH_UNIT_SEC);
    }

    calceph_interpol_PV_an(&p_pbinfile->coefftime_array, TimeJD0, Timediff, Planet, C, G, N);

    return 1;
}

/*--------------------------------------------------------------------------*/
/*!  This function computes position and velocity vectors                 
     from Chebyshev coefficients These coefficients are stored at offset  
     C, G in  the object p_pbinfile.                                      
                                          
                                                                          
 @param p_pbinfile (inout) ephemeris descriptor                            
 @param TimeJD0+timediff  (in) Time for which position is desired (Julian Date)
 @param Planet (out) Pointer to external array to receive the position/velocity.
   Planet has position in ephemerides units                                 
   Planet has velocity in ephemerides units                                 
 @param C (inout)   Coeff array entry point                                       
 @param N(in)    Number of coeff's                                           
 @param G (in)   Granules in current record                                    
*/
/*--------------------------------------------------------------------------*/
void calceph_interpol_PV_an(t_memarcoeff * p_pbinfile, treal TimeJD0,
                            treal Timediff, stateType * Planet, long C, long G, long N)
{
    treal A[15000], Cp[15000], P_Sum[3], V_Sum[3], Up[15000], T_break, T_sub, Tc;
    treal Time = TimeJD0 + Timediff;
    long long int TimeJD0pent = longr(TimeJD0);
    treal TimeJD0frac = TimeJD0 - realr(TimeJD0pent);
    long long int Timediffpent = longr(Timediff);
    treal Timedifffrac = Timediff - realr(Timediffpent);
    long long int T_begpent;
    treal T_begfrac;
    treal deltapent, deltafrac;
    int i, j;
    long int offset = 0;
    stateType X;

  /*--------------------------------------------------------------------------*/
    /* Initialize local coefficient array.                                      */
  /*--------------------------------------------------------------------------*/
    T_begpent = longr(p_pbinfile->T_beg);
    T_begfrac = p_pbinfile->T_beg - T_begpent;
    if (G == 1)
    {
        /* code equivalent a :
           Tc = 2.0E0*(Time - T_beg) / T_span - 1.0E0; */
        deltapent = (TimeJD0pent + Timediffpent - T_begpent) / p_pbinfile->T_span;
        deltafrac = (TimeJD0frac + Timedifffrac - T_begfrac) / p_pbinfile->T_span;
    }
    else
    {
        T_sub = p_pbinfile->T_span / realr(G);  /* Compute subgranule interval */

        for (j = G; j > 0; j--)
        {
            T_break = p_pbinfile->T_beg + realr(j - 1) * T_sub;
            if (Time > T_break)
            {
                /* T_seg  = T_break; */
                offset = j - 1;
                break;
            }
        }
        /* code equivalent a :
           Tc = 2.0E0*(Time - T_seg) / T_sub - 1.0E0; */
        deltapent = (TimeJD0pent + Timediffpent - T_begpent) / T_sub;
        deltafrac = (TimeJD0frac + Timedifffrac - T_begfrac) / T_sub - realr(offset);

        C = C + p_pbinfile->ncomp * offset * N;
    }
    Tc = REALC(2.0E0) * (deltapent + deltafrac) - REALC(1.0E0);
    for (i = C; i < (C + p_pbinfile->ncomp * N); i++)
        A[i - C] = p_pbinfile->Coeff_Array[i];

  /*--------------------------------------------------------------------------*/
    /* Compute the interpolated position and velocity                           */
  /*--------------------------------------------------------------------------*/

    if (p_pbinfile->ncomp == 3)
    {
        for (i = 0; i < 3; i++) /* Compute interpolating polynomials */
        {
            Cp[0] = REALC(1.0E0);
            Cp[1] = Tc;
            Cp[2] = REALC(2.0E0) * Tc * Tc - REALC(1.E0);

            Up[0] = REALC(0.0E0);
            Up[1] = REALC(1.0E0);
            Up[2] = REALC(4.0E0) * Tc;

            for (j = 3; j < N; j++)
            {
                Cp[j] = REALC(2.0E0) * Tc * Cp[j - 1] - Cp[j - 2];
                Up[j] = REALC(2.0E0) * Tc * Up[j - 1] + REALC(2.0E0) * Cp[j - 1] - Up[j - 2];
            }

            P_Sum[i] = REALC(0.0E0);    /* Compute interpolated position & velocity */
            V_Sum[i] = REALC(0.0E0);

            for (j = N - 1; j > -1; j--)
                P_Sum[i] = P_Sum[i] + A[j + i * N] * Cp[j];
            for (j = N - 1; j > 0; j--)
                V_Sum[i] = V_Sum[i] + A[j + i * N] * Up[j];

            X.Position[i] = P_Sum[i];
            X.Velocity[i] = V_Sum[i] * REALC(2.0E0) * realr(G) / (p_pbinfile->T_span);
        }
    }
    else
    {
        for (i = 0; i < 3; i++) /* Compute interpolating polynomials */
        {
            Cp[0] = REALC(1.0E0);
            Cp[1] = Tc;
            Cp[2] = REALC(2.0E0) * Tc * Tc - REALC(1.E0);

            for (j = 3; j < N; j++)
            {
                Cp[j] = REALC(2.0E0) * Tc * Cp[j - 1] - Cp[j - 2];
            }

            P_Sum[i] = REALC(0.0E0);    /* Compute interpolated position & velocity */
            V_Sum[i] = REALC(0.0E0);

            for (j = N - 1; j > -1; j--)
            {
                P_Sum[i] = P_Sum[i] + A[j + i * N] * Cp[j];
                V_Sum[i] = V_Sum[i] + A[j + (3 + i) * N] * Cp[j];
            }

            X.Position[i] = P_Sum[i];
            X.Velocity[i] = V_Sum[i];
        }
    }

  /*--------------------------------------------------------------------------*/
    /*  Return computed values.                                                 */
  /*--------------------------------------------------------------------------*/

    *Planet = X;

}

/*--------------------------------------------------------------------------*/
/*!                                                      
     This function changes the state vector, which is expressed in InputUnit units as input,  
       to the OutputUnit as output.
     
     This function is for INPOP ephemerids format.

  it returns 0 if the function fails to convert.
                                                                          
  @param pephbin (inout) ephemerids descriptor                         
  @param Target   (in) Solar system body for which position is desired.         
  @param InputUnit  (in) output unit of the statetype
  @param OutputUnit  (in) output unit of the statetype
  @param Planet (inout) position and veloticity                                                                        
          Input : expressed in InputUnit units.
          Output : expressed in OutputUnit units.
*/
/*--------------------------------------------------------------------------*/
static int calceph_inpop_unit_convert(struct calcephbin_inpop *pephbin, stateType * Planet,
                                      int Target, int InputUnit, int OutputUnit)
{
    int res = 1;
    int i;

    /* if same unit in input and output => do nothing */
    if (InputUnit == OutputUnit)
        return res;
    /* some checks */

    switch (Target)
    {
        case LIBRATIONS:       /*Librations */
            if ((OutputUnit & CALCEPH_UNIT_RAD) == 0)
                fatalerror("Units for libration must be in radians\n");
            if ((InputUnit & CALCEPH_UNIT_RAD) == 0)
                fatalerror("Input units for libration must be in radians\n");
            if (res)
                res = calceph_unit_convert_velocity_time(Planet, InputUnit, OutputUnit);
            break;

        case TCGMTCB:          /*TCG-TCB */
        case TTMTDB:           /*TT-TDB */
            if ((OutputUnit & CALCEPH_UNIT_DAY) == CALCEPH_UNIT_DAY
                && (InputUnit & CALCEPH_UNIT_SEC) == CALCEPH_UNIT_SEC)
            {
                Planet->Position[0] *= 86400E0;
            }
            if ((InputUnit & CALCEPH_UNIT_DAY) == CALCEPH_UNIT_DAY
                && (OutputUnit & CALCEPH_UNIT_SEC) == CALCEPH_UNIT_SEC)
            {
                Planet->Position[0] /= 86400E0;
            }
            if ((OutputUnit & (CALCEPH_UNIT_DAY + CALCEPH_UNIT_SEC)) == 0)
            {
                res = 0;
                fatalerror("Units must include CALCEPH_UNIT_DAY or CALCEPH_UNIT_SEC \n");
            }
            if ((OutputUnit & (CALCEPH_UNIT_DAY + CALCEPH_UNIT_SEC)) == CALCEPH_UNIT_DAY + CALCEPH_UNIT_SEC)
            {
                res = 0;
                fatalerror("Units must include only one value : CALCEPH_UNIT_DAY or CALCEPH_UNIT_SEC \n");
            }

        case NUTATIONS:
            break;

        default:
            if ((OutputUnit & (CALCEPH_UNIT_AU + CALCEPH_UNIT_KM)) == 0)
            {
                res = 0;
                fatalerror("Units must include CALCEPH_UNIT_AU or CALCEPH_UNIT_KM \n");
            }
            if ((OutputUnit & (CALCEPH_UNIT_AU + CALCEPH_UNIT_KM)) == CALCEPH_UNIT_AU + CALCEPH_UNIT_KM)
            {
                res = 0;
                fatalerror("Units must include only one value : CALCEPH_UNIT_AU or CALCEPH_UNIT_KM \n");
            }

            if ((OutputUnit & CALCEPH_UNIT_KM) == CALCEPH_UNIT_KM && (InputUnit & CALCEPH_UNIT_AU) == CALCEPH_UNIT_AU)
            {
                treal rau = pephbin->H1.AU;

                for (i = 0; i < 3; i++)
                {
                    Planet->Position[i] *= rau;
                    Planet->Velocity[i] *= rau;
                }
            }
            if ((InputUnit & CALCEPH_UNIT_KM) == CALCEPH_UNIT_KM && (OutputUnit & CALCEPH_UNIT_AU) == CALCEPH_UNIT_AU)
            {
                treal rau = pephbin->H1.AU;

                for (i = 0; i < 3; i++)
                {
                    Planet->Position[i] /= rau;
                    Planet->Velocity[i] /= rau;
                }
            }
            if (res)
                res = calceph_unit_convert_velocity_time(Planet, InputUnit, OutputUnit);
            break;
    }
    return res;
}

/*--------------------------------------------------------------------------*/
/*! return EMRAT from file 
  @param p_pbinfile (in) ephemeris file
*/
/*--------------------------------------------------------------------------*/
double calceph_inpop_getEMRAT(struct calcephbin_inpop *p_pbinfile)
{
    return p_pbinfile->H1.EMRAT;
}

/*--------------------------------------------------------------------------*/
/*! return AU from file 
  @param p_pbinfile (in) ephemeris file
*/
/*--------------------------------------------------------------------------*/
double calceph_inpop_getAU(struct calcephbin_inpop *p_pbinfile)
{
    return p_pbinfile->H1.AU;
}

/*--------------------------------------------------------------------------*/
/*! return the value of the specified constant from file 
  return 0 if constant is not found
  return 1 if constant is found and set p_pdval.
  @param p_pbinfile (in) ephemeris file
  @param name (in) constant name
  @param p_pdval (out) value of the constant
*/
/*--------------------------------------------------------------------------*/
int calceph_inpop_getconstant(struct calcephbin_inpop *p_pbinfile, const char *name, double *p_pdval)
{
    int found = 0;
    int cpt = 0;
    size_t len = strlen(name);

    while (cpt < 400 && found == 0)
    {
        if (strncmp(name, (const char *) (p_pbinfile->H1.constName[cpt]), len) == 0)
        {
            found = 1;
            *p_pdval = p_pbinfile->constVal[cpt];
        }
        cpt++;
    }
    return found;
}

/*--------------------------------------------------------------------------*/
/*! return the number of constants available in the ephemeris file 
  
  @param eph (inout) ephemeris descriptor
*/
/*--------------------------------------------------------------------------*/
int calceph_inpop_getconstantcount(struct calcephbin_inpop *eph)
{
    return eph->H1.numConst + 2;    /*+2 <=> AU + EMRAT */
}

/*--------------------------------------------------------------------------*/
/*! get the name and its value of the constant available 
    at some index from the ephemeris file 
   return 1 if index is valid
   return 0 if index isn't valid
  
  @param eph (inout) ephemeris descriptor
  @param index (in) index of the constant (must be between 1 and calceph_getconstantcount() )
  @param name (out) name of the constant 
  @param value (out) value of the constant 
*/
/*--------------------------------------------------------------------------*/
int calceph_inpop_getconstantindex(struct calcephbin_inpop *eph, int index, char name[CALCEPH_MAX_CONSTANTNAME],
                                   double *value)
{
    int res;
    int j;
    int nconst = calceph_inpop_getconstantcount(eph);

    if (index >= 1 && index <= nconst)
    {
        res = 1;
        for (j = 0; j < CALCEPH_MAX_CONSTANTNAME; j++)
            name[j] = ' ';      /* initialization for fortran */
        if (index == nconst)
        {
            *value = calceph_inpop_getAU(eph);
            strcpy(name, "AU");
        }
        else if (index == nconst - 1)
        {
            *value = calceph_inpop_getEMRAT(eph);
            strcpy(name, "EMRAT");
        }
        else
        {
            strncpy(name, eph->H1.constName[index - 1], 6 /*CALCEPH_MAX_CONSTANTNAME-1 */ );
            *value = eph->constVal[index - 1];
        }
        name[6 /*CALCEPH_MAX_CONSTANTNAME-1 */ ] = '\0';
    }
    else
    {
        res = 0;
    }
    return res;
}
