/*-----------------------------------------------------------------*/
/*! 
  \file calcephspkio.c 
  \brief perform I/O on SPICE KERNEL ephemeris data file 
         interpolate SPICE KERNEL Ephemeris data.
         compute position and velocity from SPICE KERNEL ephemeris file.

  \author  M. Gastineau 
           Astronomie et Systemes Dynamiques, IMCCE, CNRS, Observatoire de Paris. 

   Copyright, 2011, 2012, CNRS
   email of the author : gastineau@imcce.fr

  History:                                                                
*/
/*-----------------------------------------------------------------*/

/*-----------------------------------------------------------------*/
/* License  of this file :
 This file is "triple-licensed", you have to choose one  of the three licenses 
 below to apply on this file.
 
    CeCILL-C
    	The CeCILL-C license is close to the GNU LGPL.
    	( http://www.cecill.info/licences/Licence_CeCILL-C_V1-en.html )
   
 or CeCILL-B
        The CeCILL-B license is close to the BSD.
        (http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.txt)
  
 or CeCILL v2.0
      The CeCILL license is compatible with the GNU GPL.
      ( http://www.cecill.info/licences/Licence_CeCILL_V2-en.html )
 

This library is governed by the CeCILL-C, CeCILL-B or the CeCILL license under 
French law and abiding by the rules of distribution of free software.  
You can  use, modify and/ or redistribute the software under the terms 
of the CeCILL-C,CeCILL-B or CeCILL license as circulated by CEA, CNRS and INRIA  
at the following URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C,CeCILL-B or CeCILL license and that you accept its terms.
*/
/*-----------------------------------------------------------------*/

#include "calcephconfig.h"
#if HAVE_STDIO_H
#include <stdio.h>
#endif
#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if HAVE_STRING_H
#include <string.h>
#endif
#if HAVE_ERRNO_H
#include <errno.h>
#endif
#if HAVE_MATH_H
#include <math.h>
#endif
#if HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifndef __cplusplus
#if HAVE_STDBOOL_H
#include <stdbool.h>
#endif
#endif /*__cplusplus */

#include "calcephdebug.h"
#include "real.h"
#define __CALCEPH_WITHIN_CALCEPH 1
#include "calceph.h"
#include "calcephspice.h"
#include "util.h"

/*--------------------------------------------------------------------------*/
/* private functions */
/*--------------------------------------------------------------------------*/
#if DEBUG
static void calceph_spk_debugsegment(struct SPKSegmentHeader *seg);
#endif
static int calceph_spk_readlistsegment(FILE * file, const char *filename, struct SPKSegmentList **list, int fwd,
                                       int bwd);
static int calceph_spk_readsegment_header(FILE * file, const char *filename, struct SPKSegmentHeader *seg);
static void calceph_spk_interpol_PV_an(const double *Coeff_Array, treal TimeJD0,
                                       treal Timediff, stateType * Planet, int N, int ncomp);

/*--------------------------------------------------------------------------*/
/*!  
     Read the header data block for the SPK ephemeris file
     Read the comment block.
     Read the segment block.
     
  @return  1 on sucess and 0 on failure
  
  @param file (inout) file descriptor.   
  @param filename (in) A character string giving the name of an ephemeris data file.   
  @param res (out) descriptor of the ephemeris.  
*/
/*--------------------------------------------------------------------------*/
int calceph_spk_open(FILE * file, const char *filename, struct SPICEkernel *res)
{
    struct SPKHeader header;
    struct SPKSegmentList *listseg;

    listseg = NULL;
 /*---------------------------------------------------------*/
    /* read the header block */
 /*---------------------------------------------------------*/
    /* go to the head of the file */
    if (fseeko(file, 0, SEEK_SET) != 0)
    {
        fatalerror("Can't jump to the beginning of the ephemeris file '%s'\nSystem error : '%s'\n",
                   filename, strerror(errno));
        return 0;
    }
    if (fread(&header, sizeof(header), 1, file) != 1)
    {
        fatalerror("Can't read the beginning of the ephemeris file '%s'\nSystem error : '%s'\n",
                   filename, strerror(errno));
        return 0;
    }

#if DEBUG
    calceph_spk_debugheader(&header);
#endif
 /*---------------------------------------------------------*/
    /* check the ftp string */
 /*---------------------------------------------------------*/
    if (calceph_spk_ftp(header.ftp) == 0)
    {
        fatalerror("The FTP string is not valid in the file '%s'\n", filename);
        return 0;
    }
 /*---------------------------------------------------------*/
    /* read the list of segments */
 /*---------------------------------------------------------*/
    if (calceph_spk_readlistsegment(file, filename, &listseg, header.fwd, header.bwd) == 0)
    {
        return 0;
    }
    res->filetype = DAF_SPK;
    res->filedata.spk.header = header;
    res->filedata.spk.list_seg = listseg;
    res->filedata.spk.file = file;
    return 1;
}

/*--------------------------------------------------------------------------*/
/*! Close the SPICE ephemeris file.
     
  @return  1 on sucess and 0 on failure
  
  @param eph (inout) descriptor of the ephemeris.  
*/
/*--------------------------------------------------------------------------*/
void calceph_spk_close(struct SPKfile *eph)
{
    while (eph->list_seg != NULL)
    {
        struct SPKSegmentList *listseg = eph->list_seg;

        eph->list_seg = eph->list_seg->next;
        free(listseg);
    }
    fclose(eph->file);
}

/*--------------------------------------------------------------------------*/
/*!  
     Debug the header of a SPK file
  
  @param header (in) header of SPK
*/
/*--------------------------------------------------------------------------*/
#if DEBUG
void calceph_spk_debugheader(const struct SPKHeader *header)
{
    int k;

    printf("file architecture and type : '%*s'\n", IDWORD_LEN, header->idword);
    printf("ND : %d\n", header->nd);
    printf("NI : %d\n", header->ni);
    printf("Internal file name : '%*s'\n", IFNAME_LEN, header->ifname);
    printf("FWD : %d\n", header->fwd);
    printf("BWD : %d\n", header->bwd);
    printf("FREE : %d\n", header->free);
    printf("Binary file format ID : '%*s'\n", IFNAME_LEN, header->bff);
    printf("FTP corruption test string : \n");
    for (k = 0; k < FTP_LEN; k++)
        printf("%d ", (int) header->ftp[k]);
    printf("\n");
}
#endif

/*--------------------------------------------------------------------------*/
/*! return 0 if fptbuf is not equal to the string
  FTPSTR:<13>:<10>:<13><10>:<13><0>:<129>:<16><206>:ENDFTP
  
  return 1 if ftp is esqual to this string
  
  @param ftpbuf (in) buffer to checked
*/
/*--------------------------------------------------------------------------*/
int calceph_spk_ftp(const unsigned char ftpbuf[FTP_LEN])
{
    const unsigned char validftp[FTP_LEN] =
        { 'F', 'T', 'P', 'S', 'T', 'R', ':', 13, ':', 10, ':', 13, 10, ':', 13, 0, ':', 129, ':', 16, 206, ':', 'E',
'N', 'D', 'F', 'T', 'P' };
    int j;
    int res = 1;

    for (j = 0; j < FTP_LEN; j++)
    {
        if (ftpbuf[j] != validftp[j])
            res = 0;
    }
    return res;
}

/*--------------------------------------------------------------------------*/
/*! read the list of segments in the file : only the header                 

   @param file (inout) file descriptor.   
   @param filename (in) A character string giving the name of an ephemeris data file.   
   @param list (inout) pointer to the list 
   @param fwd (in) record number to read
   @param bwd (in) previous record number
*/
/*--------------------------------------------------------------------------*/
int calceph_spk_readlistsegment(FILE * file, const char *filename, struct SPKSegmentList **list, int fwd, int bwd)
{
    char descriptor[DAF_RECORD_LEN];
    double *ddescriptor = (double *) descriptor;
    char segmentid[DAF_RECORD_LEN];
    int next = 0, count;
#if DEBUG
    int prev;
#endif
   int res = 1;
    int j, offset;
    struct SPKSegmentList *pnew;

#if DEBUG
    printf("calceph_spk_readlistsegment(%p,%d)\n", *list, fwd);
#endif

    /* go to the head of the file */
    if (fseeko(file, DAF_RECORD_LEN * (fwd - 1), SEEK_SET) != 0)
    {
        fatalerror
            ("Can't jump to the segment descriptor record at location %d of the ephemeris file '%s'\nSystem error : '%s'\n",
             fwd, filename, strerror(errno));
        return 0;
    }
    if (fread(descriptor, sizeof(char), DAF_RECORD_LEN, file) != DAF_RECORD_LEN)
    {
        fatalerror
            ("Can't read the segment descriptor record at location  %d of the ephemeris file '%s'\nSystem error : '%s'\n",
             fwd, filename, strerror(errno));
        return 0;
    }
    if (fread(segmentid, sizeof(char), DAF_RECORD_LEN, file) != DAF_RECORD_LEN)
    {
        fatalerror("Can't read the segment id record at location  %d of the ephemeris file '%s'\nSystem error : '%s'\n",
                   fwd + 1, filename, strerror(errno));
        return 0;
    }
    next = (int) ddescriptor[0];
#if DEBUG
    prev = (int) ddescriptor[1];
#endif
    count = (int) ddescriptor[2];

#if DEBUG
    printf("iprev=%d %d %d\n", prev, next, count);

    for (j = 0; j < DAF_RECORD_LEN / 8; j++)
        printf("%g ", ((double *) descriptor)[j]);
    printf("\n");
    for (j = 0; j < DAF_RECORD_LEN; j++)
        printf("%c ", segmentid[j]);
    printf("\n");
#endif

    /* create a new element of the list */
    pnew = (struct SPKSegmentList *) malloc(sizeof(struct SPKSegmentList));
    if (pnew == NULL)
    {
        fatalerror("Can't allocate memory for a segment descritor\n");
        return 0;
    }
    /* insert in the list */
    pnew->next = *list;
    pnew->prev = NULL;
    if (*list != NULL)
        (*list)->prev = pnew;
    *list = pnew;
    pnew->recordnumber = fwd;
    pnew->array_seg.count = count;

    /* fill the content of the list */
    offset = 2;
    for (j = 0; j < count && res == 1; j++)
    {
        const int *parint = (int *) (ddescriptor + offset + 3);
        struct SPKSegmentHeader *seg = pnew->array_seg.array + j;

        seg->T_begin = ddescriptor[offset + 1];
        seg->T_end = ddescriptor[offset + 2];
        seg->body = parint[0];
        seg->center = parint[1];
        seg->refframe = parint[2];
        seg->datatype = (enum SPKdatatype) parint[3];
        seg->rec_begin = parint[4];
        seg->rec_end = parint[5];

        memcpy(seg->id, segmentid + SEGMENTID_LEN * j, SEGMENTID_LEN);
        memcpy(seg->descriptor, ddescriptor + offset, NS_SPK * 8);
#if DEBUG
        printf("before calceph_spk_readsegment_header:\n");
        calceph_spk_debugsegment(seg);
#endif

        res = calceph_spk_readsegment_header(file, filename, seg);
#if DEBUG
        printf("after calceph_spk_readsegment_header:\n");
        calceph_spk_debugsegment(seg);
#endif
        offset += NS_SPK;
    }

    /* go the next records */
    if (next != 0 && res != 0)
    {
        res = calceph_spk_readlistsegment(file, filename, list, next, fwd);
    }
    return res;
}

/*--------------------------------------------------------------------------*/
/* ! read the double-precision numbers from rec_begin to rec_end (in number of double-precision numbers)
   @param file (inout) file descriptor.   
   @param filename (in) A character string giving the name of an ephemeris data file.   
   @param rec_begin (inout) first word to read 
   @param rec_end (inout) last word to read 
   @param record (out) array to store the words
*/
/*--------------------------------------------------------------------------*/
int calceph_spk_readword(FILE * file, const char *filename, int rec_begin, int rec_end, double *record)
{
    int n = rec_end - rec_begin + 1;

    if (fseeko(file, (rec_begin - 1) * sizeof(double), SEEK_SET) != 0)
    {
        fatalerror("Can't jump to the location %lu of the ephemeris file '%s'\nSystem error : '%s'\n",
                   rec_begin, filename, strerror(errno));
        return 0;
    }
    if (fread(record, sizeof(double), n, file) != n)
    {
        fatalerror("Can't read %d data at location  %d of the ephemeris file '%s'\nSystem error : '%s'\n",
                   n, rec_begin, filename, strerror(errno));
        return 0;
    }

    return 1;
}

/*--------------------------------------------------------------------------*/
/* ! debug the segment 
   @param seg (inout) segment 
*/
/*--------------------------------------------------------------------------*/
#if DEBUG
void calceph_spk_debugsegment(struct SPKSegmentHeader *seg)
{
    printf("segment ID : '%.*s'\n", SEGMENTID_LEN, seg->id);
    printf("T_begin : %e -> %e\n", seg->T_begin, seg->T_end);
    printf("body : %d \n", seg->body);
    printf("center : %d \n", seg->center);
    printf("refframe : %d \n", seg->refframe);
    printf("datatype : %d \n", (int) seg->datatype);
    printf("rec_begin : %d \n", seg->rec_begin);
    printf("rec_end : %d \n", seg->rec_end);
    switch (seg->datatype)
    {
        case SPK_SEGTYPE3:
        case SPK_SEGTYPE2:
            printf("information of the segment : %e %e %d %d\n",
                   seg->seginfo.data2.T_begin, seg->seginfo.data2.T_len,
                   seg->seginfo.data2.count_dataperrecord, seg->seginfo.data2.count_record);
            break;

        default:
            printf("Unsupported segment (type=%d).", (int) seg->datatype);
            break;
    }
}
#endif

/*--------------------------------------------------------------------------*/
/* ! read the data header of thios segmentid
   @param file (inout) file descriptor.   
   @param filename (in) A character string giving the name of an ephemeris data file.   
   @param seg (inout) segment 
*/
/*--------------------------------------------------------------------------*/
int calceph_spk_readsegment_header(FILE * file, const char *filename, struct SPKSegmentHeader *seg)
{
    double drecord[6];
    int res = 1;

    switch (seg->datatype)
    {
        case SPK_SEGTYPE3:
        case SPK_SEGTYPE2:
            res = calceph_spk_readword(file, filename, seg->rec_end - 3, seg->rec_end, drecord);
            if (res == 1)
            {
                seg->seginfo.data2.T_begin = drecord[0];
                seg->seginfo.data2.T_len = drecord[1];
                seg->seginfo.data2.count_dataperrecord = (int) drecord[2];
                seg->seginfo.data2.count_record = (int) drecord[3];
            }
            break;

        default:
            fatalerror("Unsupported segment (type=%d).", (int) seg->datatype);
            res = 0;
            break;
    }
    return res;
}

/*--------------------------------------------------------------------------*/
/*!                                                      
     This function read the data block corresponding to the specified Time
     if this block isn't already in memory.
     
     This function works only on all targets.
     for asteroids, the target value is "asteroid number"+CALCEPH_ASTEROID
     
     It computes position and velocity vectors for a selected  
     planetary body from Chebyshev coefficients.          

     return 1 on success
     return 0 on error  
     return -1 if the segment and object is not found
                                                                         
                                                                          
  @param eph (inout) ephemerids descriptor                         
  @param pspk (inout) SPK file                         
  @param TimeJD0+timediff  (in) Time for which position is desired (Julian Date)
  @param Target   (in) Solar system body for which position is desired.         
  @param Center   (in) Solar system body from which position is desired.         
  @param Planet (out) position and veloticity                                                                        
         Planet has position in ua                                                
         Planet has velocity in ua/day
*/
/*--------------------------------------------------------------------------*/
int calceph_spk_interpol_PV(struct calcephbin_spice *eph, struct SPKfile *pspk,
                            treal TimeJD0, treal Timediff, int Target, int Center, stateType * Planet)
{
    double Time = TimeJD0 + Timediff;
    int res = -1;
    int j;
    int spicetarget = calceph_spice_convertid(eph, Target);
    int spicecenter = calceph_spice_convertid(eph, Center);
    struct SPKSegmentList * listseg;
    
    if (spicetarget == -1)
    {
        fatalerror("Target object %d is not supported.\n", Target);
        return 0;
    }
    if (spicecenter == -1)
    {
        fatalerror("Center object %d is not supported.\n", Center);
        return 0;
    }

    /* find the segment where is located this time */
    for (listseg = pspk->list_seg; listseg != NULL && res == -1; listseg = listseg->next)
    {
        for (j = 0; j < listseg->array_seg.count && res == -1; j++)
        {
            struct SPKSegmentHeader *seg = listseg->array_seg.array + j;

#if DEBUG
            printf("target: %d %d %d %d %d %d\n", seg->body, seg->center, spicetarget, spicecenter, Target, Center);
#endif       /*DEBUG*/
                if (seg->body != spicetarget)
                continue;
            if (seg->center != spicecenter)
                continue;
            if (seg->refframe != 1)
                continue;

#if DEBUG
            printf("%23.16E %23.16E %23.16E\n", seg->T_begin / 86400E0 + 2.451545E+06, Time,
                   seg->T_end / 86400E0 + 2.451545E+06);
#endif       /*DEBUG*/
                if (seg->T_begin / 86400E0 + 2.451545E+06 > Time || seg->T_end / 86400E0 + 2.451545E+06 < Time)
                continue;

#if DEBUG
            printf("Find correct segment\n");
#endif       /*DEBUG*/
                /* we found the valid segment */
                switch (seg->datatype)
            {
                case SPK_SEGTYPE3:
                case SPK_SEGTYPE2:
                    res = calceph_spk_interpol_PV_segment_2(pspk, seg, TimeJD0, Timediff, Planet);
                    break;

                default:
                    //fatalerror("Unsupported segment (type=%d).", (int)seg->datatype);
                    break;
            }

        }
        // printf("\n");
    }

#if DEBUG
    if (res)
    {
        printf("answer for %d %d %23.16E : \n", Target, Center, Time);
        for (j = 0; j <= 2; j++)
            printf("%g ", Planet->Position[j]);
        puts("\n");
        for (j = 0; j <= 2; j++)
            printf("%g ", Planet->Velocity[j]);
        puts("\n");
    }
#endif
    return res;
}

/*--------------------------------------------------------------------------*/
/*!                                                      
     This function read the data block corresponding to the specified Time
     if this block isn't already in memory.
     
     This function works only over the specified segment.
     The segment must must of type 2 or 3.
     
     It computes position and velocity vectors for a selected  
     planetary body from Chebyshev coefficients.          

                                                                          
  @param pspk (inout) SPK file  
  @param seg (in) segment where is located TimeJD0+timediff 
  @param TimeJD0+timediff  (in) Time for which position is desired (Julian Date)
  @param Planet (out) position and veloticity                                                                        
         Planet has position in ua                                                
         Planet has velocity in ua/day
*/
/*--------------------------------------------------------------------------*/
int calceph_spk_interpol_PV_segment_2(struct SPKfile *pspk, struct SPKSegmentHeader *seg, treal TimeJD0, treal Timediff,
                                      stateType * Planet)
{
    const double Timesec = ((TimeJD0 - 2.451545E+06) + Timediff) * 86400E0;
    int recT;                   /* record where is located Timesec */
    double drecord[1000];
    int wordT;
    const int nword = seg->seginfo.data2.count_dataperrecord;

#if DEBUG
    int j;
#endif
    int ncomp, N;

    recT = (Timesec - seg->seginfo.data2.T_begin) / seg->seginfo.data2.T_len;
    if (recT == seg->seginfo.data2.count_record &&
        Timesec <= seg->seginfo.data2.T_begin + seg->seginfo.data2.T_len * recT)
        recT = recT - 1;

    if (recT < 0 || recT >= seg->seginfo.data2.count_record)
    {
        fatalerror
            ("Computation of record is not valid for segment of type 2. Looking for time %23.16E. Beginning time in segment : %23.16E\n"
             "Time slice in the segment : %23.16E\n. Number of records: %d\n Coumputed record : %d\n", Timesec,
             seg->seginfo.data2.T_begin, seg->seginfo.data2.T_len, seg->seginfo.data2.count_record, recT);
        return 0;
    }

    /* read the data */
    wordT = (seg->rec_begin + recT * nword);
    if (calceph_spk_readword(pspk->file, "", wordT, wordT + nword - 1, drecord) == 0)
    {
        return 0;
    }
    /*for(j=0; j<nword; j++) printf("%23.16E\n", drecord[j]); */
#if DEBUG
    printf
        ("seg->seginfo.data2.T_begin=%23.16E seg->seginfo.data2.T_len=%23.16E begin record=%23.16E end record =%23.16E Timesec=%23.16E\n",
         seg->seginfo.data2.T_begin - seg->seginfo.data2.T_len, seg->seginfo.data2.T_begin + seg->seginfo.data2.T_len,
         drecord[0], drecord[0] + drecord[1], Timesec);
#endif

    /* computation */
    ncomp = (seg->datatype == SPK_SEGTYPE3) ? 6 : 3;
    N = (nword - 2) / ncomp;
    /*printf("ncomp=%d %d\n", ncomp, N); */

    calceph_spk_interpol_PV_an(drecord, (TimeJD0 - 2.451545E+06) * 86400E0, Timediff * 86400E0, Planet, N, ncomp);

#if DEBUG
    for (j = 0; j < 3; j++)
        printf("%23.16E %23.16E\n", Planet->Position[j], Planet->Velocity[j]);
    printf("%23.16E\n",
           sqrt(Planet->Position[0] * Planet->Position[0] + Planet->Position[1] * Planet->Position[1] +
                Planet->Position[2] * Planet->Position[2]));
#endif

    return 1;
}

/*--------------------------------------------------------------------------*/
/*!  This function computes position and velocity vectors                 
     from Chebyshev coefficients from Coeff_Array.                                      
                                          
                                                                          
 @param Coeff_Array (in) array of Tchebychev coefficient                         
 @param TimeJD0+timediff  (in) Time for which position is desired (Julian Date)
 @param Planet (out) Pointer to external array to receive the position/velocity.
   Planet has position in ephemerides units                                 
   Planet has velocity in ephemerides units                                 
 @param N(in)    Number of coeff's      
 @param ncomp (in) number of component
*/
/*--------------------------------------------------------------------------*/
static void calceph_spk_interpol_PV_an(const double *Coeff_Array, treal TimeJD0,
                                       treal Timediff, stateType * Planet, int N, int ncomp)
{
    treal Cp[15000], P_Sum[3], V_Sum[3], Up[15000], Tc;
    long long int TimeJD0pent = longr(TimeJD0);
    treal TimeJD0frac = TimeJD0 - realr(TimeJD0pent);
    long long int Timediffpent = longr(Timediff);
    treal Timedifffrac = Timediff - realr(Timediffpent);
    long long int T_begpent;
    treal T_begfrac;
    treal deltapent, deltafrac;
    int i, j;
    stateType X;
    const double T_beg = Coeff_Array[0] - Coeff_Array[1];
    const double T_span = Coeff_Array[1] * 2.E0;
    const double *A = Coeff_Array + 2;

  /*--------------------------------------------------------------------------*/
    /* Compute Tc                           */
  /*--------------------------------------------------------------------------*/

    T_begpent = longr(T_beg);
    T_begfrac = T_beg - T_begpent;
    /* code equivalent a :
       Tc = 2.0E0*(Time - T_beg) / T_span - 1.0E0; */
    deltapent = (TimeJD0pent + Timediffpent - T_begpent) / T_span;
    deltafrac = (TimeJD0frac + Timedifffrac - T_begfrac) / T_span;
    Tc = REALC(2.0E0) * (deltapent + deltafrac) - REALC(1.0E0);

  /*--------------------------------------------------------------------------*/
    /* Compute the interpolated position and velocity                           */
  /*--------------------------------------------------------------------------*/

    if (ncomp == 3)
    {
        for (i = 0; i < 3; i++) /* Compute interpolating polynomials */
        {
            Cp[0] = REALC(1.0E0);
            Cp[1] = Tc;
            Cp[2] = REALC(2.0E0) * Tc * Tc - REALC(1.E0);

            Up[0] = REALC(0.0E0);
            Up[1] = REALC(1.0E0);
            Up[2] = REALC(4.0E0) * Tc;

            for (j = 3; j < N; j++)
            {
                Cp[j] = REALC(2.0E0) * Tc * Cp[j - 1] - Cp[j - 2];
                Up[j] = REALC(2.0E0) * Tc * Up[j - 1] + REALC(2.0E0) * Cp[j - 1] - Up[j - 2];
            }

            P_Sum[i] = REALC(0.0E0);    /* Compute interpolated position & velocity */
            V_Sum[i] = REALC(0.0E0);

            for (j = N - 1; j > -1; j--)
                P_Sum[i] = P_Sum[i] + A[j + i * N] * Cp[j];
            for (j = N - 1; j > 0; j--)
                V_Sum[i] = V_Sum[i] + A[j + i * N] * Up[j];

            X.Position[i] = P_Sum[i];
            X.Velocity[i] = V_Sum[i] * REALC(2.0E0) / (T_span);
        }
    }
    else
    {
        for (i = 0; i < 3; i++) /* Compute interpolating polynomials */
        {
            Cp[0] = REALC(1.0E0);
            Cp[1] = Tc;
            Cp[2] = REALC(2.0E0) * Tc * Tc - REALC(1.E0);

            for (j = 3; j < N; j++)
            {
                Cp[j] = REALC(2.0E0) * Tc * Cp[j - 1] - Cp[j - 2];
            }

            P_Sum[i] = REALC(0.0E0);    /* Compute interpolated position & velocity */
            V_Sum[i] = REALC(0.0E0);

            for (j = N - 1; j > -1; j--)
            {
                P_Sum[i] = P_Sum[i] + A[j + i * N] * Cp[j];
                V_Sum[i] = V_Sum[i] + A[j + (3 + i) * N] * Cp[j];
            }

            X.Position[i] = P_Sum[i];
            X.Velocity[i] = V_Sum[i];
        }
    }

  /*--------------------------------------------------------------------------*/
    /*  Return computed values.                                                 */
  /*--------------------------------------------------------------------------*/

    *Planet = X;

}
