!/*-----------------------------------------------------------------*/
!/*! 
!  \file f2003calceph.f 
!  \brief fortran 2003 binding interface
!         requires a fortran compiler with the standard fortran 2003.
!
!  \author  M. Gastineau 
!           Astronomie et Systemes Dynamiques, IMCCE, CNRS, Observatoire de Paris. 
!
!   Copyright, 2008, 2009, 2010, 2011, CNRS
!   email of the author : gastineau@imcce.fr
! 
! History :
!    \note M. Gastineau 12/10/10 : add custom error handler     
!    \note M. Gastineau 21/05/11 : support asteroid in INPOP file format 2.0
!    \note M. Gastineau 10/06/11 : support SPICE kernel file
!    \note M. Gastineau 22/08/11 : add calceph_compute_unit
!    \note M. Gastineau 15/12/11 : wrong interface in/out for calceph_sgetconstantindex
!
!*/
!/*-----------------------------------------------------------------*/

!/*-----------------------------------------------------------------*/
!/* License  of this file :
!  This file is "triple-licensed", you have to choose one  of the three licenses 
!  below to apply on this file.
!  
!     CeCILL-C
!     	The CeCILL-C license is close to the GNU LGPL.
!     	( http://www.cecill.info/licences/Licence_CeCILL-C_V1-en.html )
!   
!  or CeCILL-B
!        The CeCILL-B license is close to the BSD.
!        (http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.txt)
!  
!  or CeCILL v2.0
!       The CeCILL license is compatible with the GNU GPL.
!       ( http://www.cecill.info/licences/Licence_CeCILL_V2-en.html )
!  
! 
! This library is governed by the CeCILL-C, CeCILL-B or the CeCILL license under 
! French law and abiding by the rules of distribution of free software.  
! You can  use, modify and/ or redistribute the software under the terms 
! of the CeCILL-C,CeCILL-B or CeCILL license as circulated by CEA, CNRS and INRIA  
! at the following URL "http://www.cecill.info". 
!
! As a counterpart to the access to the source code and  rights to copy,
! modify and redistribute granted by the license, users are provided only
! with a limited warranty  and the software's author,  the holder of the
! economic rights,  and the successive licensors  have only  limited
! liability. 
!
! In this respect, the user's attention is drawn to the risks associated
! with loading,  using,  modifying and/or developing or reproducing the
! software by the user in light of its specific status of free software,
! that may mean  that it is complicated to manipulate,  and  that  also
! therefore means  that it is reserved for developers  and  experienced
! professionals having in-depth computer knowledge. Users are therefore
! encouraged to load and test the software's suitability as regards their
! requirements in conditions enabling the security of their systems and/or 
! data to be ensured and,  more generally, to use and operate it in the 
! same conditions as regards security. 
!
! The fact that you are presently reading this means that you have had
! knowledge of the CeCILL-C,CeCILL-B or CeCILL license and that you accept its terms.
! */
! /*-----------------------------------------------------------------*/

        MODULE calceph
!/*----------------------------------------------------------------------------------------------*/
!/* definition of some constants */
!/*----------------------------------------------------------------------------------------------*/
!/*! define the maximum number of characters (including the trailing '\0') 
! that the name of a constant could contain. */
        INTEGER, parameter :: CALCEPH_MAX_CONSTANTNAME=33

!/*! define the offset value for asteroid for calceph_?compute */
        INTEGER, parameter :: CALCEPH_ASTEROID=2000000

!/* unit for the output */
!/*! outputs are in Astronomical Unit */
        INTEGER, parameter :: CALCEPH_UNIT_AU=1
!/*! outputs are in kilometers */
        INTEGER, parameter :: CALCEPH_UNIT_KM=2
!/*! outputs are in day */
        INTEGER, parameter :: CALCEPH_UNIT_DAY=4
!/*! outputs are in seconds */
        INTEGER, parameter :: CALCEPH_UNIT_SEC=8
!/*! outputs are in radians */
        INTEGER, parameter :: CALCEPH_UNIT_RAD=16

        INTERFACE
 
!/*-----------------------------------------------------------------*/
!/* error handler */
!/*-----------------------------------------------------------------*/
        ! set the error handler 
        SUBROUTINE calceph_seterrorhandler(typeh,userfunch) BIND(C)
          USE, INTRINSIC :: ISO_C_BINDING
          IMPLICIT NONE
          integer(C_INT), VALUE, intent(in) :: typeh
          type(c_funptr), VALUE, intent(in) :: userfunch
        END SUBROUTINE calceph_seterrorhandler

!/*-----------------------------------------------------------------*/
!/* single access API per thread/process */
!/*-----------------------------------------------------------------*/
        ! open an ephemeris data file 
        FUNCTION calceph_sopen(filename) BIND(C)
          USE, INTRINSIC :: ISO_C_BINDING
          IMPLICIT NONE
          character(len=1,kind=C_CHAR), intent(in) :: filename
          integer(C_INT) :: calceph_sopen
        END FUNCTION calceph_sopen


        ! compute the position <x,y,z> and velocity <xdot,ydot,zdot>
        !    for a given target and center 
       FUNCTION calceph_scompute(JD0,time,target1,center,PV) BIND(C)
          USE, INTRINSIC :: ISO_C_BINDING
          IMPLICIT NONE
          INTEGER(C_INT), VALUE, intent(in) :: target1, center
          REAL(C_DOUBLE), VALUE, intent(in) :: JD0, time
          REAL(C_DOUBLE), intent(out) :: PV(6)
          INTEGER(C_INT) :: calceph_scompute
       END FUNCTION calceph_scompute
   
        ! get constant value from the specified name
        FUNCTION calceph_sgetconstant(name, value) BIND(C)
          USE, INTRINSIC :: ISO_C_BINDING
          IMPLICIT NONE
          character(len=1,kind=C_CHAR), intent(in) :: name
          REAL(C_DOUBLE), intent(out) :: value
          integer(C_INT) :: calceph_sgetconstant
        END FUNCTION calceph_sgetconstant

        ! get the number of constants available in the ephemeris file
        FUNCTION calceph_sgetconstantcount() BIND(C)
          USE, INTRINSIC :: ISO_C_BINDING
          IMPLICIT NONE
          integer(C_INT) :: calceph_sgetconstantcount
        END FUNCTION calceph_sgetconstantcount

        ! get the name and its value of the constant available at some index in the ephemeris file
        FUNCTION calceph_sgetconstantindex(index, name, value) BIND(C)
          USE, INTRINSIC :: ISO_C_BINDING
          IMPLICIT NONE
          integer(C_INT), VALUE, intent(in) :: index
          character(len=1,kind=C_CHAR),dimension(33),intent(out)::name
          REAL(C_DOUBLE), intent(out) :: value
          integer(C_INT) :: calceph_sgetconstantindex
        END FUNCTION calceph_sgetconstantindex

        ! close an ephemeris data file 
       SUBROUTINE calceph_sclose() BIND(C)
          USE, INTRINSIC :: ISO_C_BINDING
          IMPLICIT NONE
       END SUBROUTINE calceph_sclose

!/*-----------------------------------------------------------------*/
!/* multiple access API per thread/process */
!/*-----------------------------------------------------------------*/

        ! open an ephemeris data file
        FUNCTION calceph_open(filename) BIND(C)
          USE, INTRINSIC :: ISO_C_BINDING
          IMPLICIT NONE
          CHARACTER(len=1,kind=C_CHAR), intent(in) :: filename
          TYPE(C_PTR) :: calceph_open
        END FUNCTION calceph_open

        ! open several ephemeris data files
        FUNCTION calceph_open_array(n, array_filename, lenfilename)      &
     &             BIND(C,  name="f2003calceph_open_array")
          USE, INTRINSIC :: ISO_C_BINDING
          IMPLICIT NONE
          INTEGER(C_INT), VALUE, intent(in) :: n
          CHARACTER(len=1,kind=C_CHAR),dimension(*), intent(in) ::       &
     &             array_filename
          INTEGER(C_INT), VALUE, intent(in) :: lenfilename
          TYPE(C_PTR) :: calceph_open_array
        END FUNCTION calceph_open_array

       ! compute the position <x,y,z> and velocity <xdot,ydot,zdot>
       !    for a given target and center (outputs are expressed in UA and UA/day)
       FUNCTION calceph_compute(eph,JD0,time,target1,                    &
     &                             center,PV) BIND(C)
          USE, INTRINSIC :: ISO_C_BINDING
          IMPLICIT NONE
          INTEGER(C_INT), VALUE, intent(in) :: target1, center
          REAL(C_DOUBLE), VALUE, intent(in) :: JD0, time
          REAL(C_DOUBLE), intent(out) :: PV(6)
          TYPE(C_PTR), VALUE, intent(in) :: eph
          INTEGER(C_INT) :: calceph_compute
       END FUNCTION calceph_compute

       ! compute the position <x,y,z> and velocity <xdot,ydot,zdot>
       !    for a given target and center (outputs are expressed in unit)
       FUNCTION calceph_compute_unit(eph,JD0,time,target1,               &
     &                             center,unit, PV) BIND(C)
          USE, INTRINSIC :: ISO_C_BINDING
          IMPLICIT NONE
          INTEGER(C_INT), VALUE, intent(in) :: target1, center
          INTEGER(C_INT), VALUE, intent(in) :: unit
          REAL(C_DOUBLE), VALUE, intent(in) :: JD0, time
          REAL(C_DOUBLE), intent(out) :: PV(6)
          TYPE(C_PTR), VALUE, intent(in) :: eph
          INTEGER(C_INT) :: calceph_compute_unit
       END FUNCTION calceph_compute_unit

        ! get constant value from the specified name
        FUNCTION calceph_getconstant(eph, name, value) BIND(C)
          USE, INTRINSIC :: ISO_C_BINDING
          IMPLICIT NONE
          CHARACTER(len=1,kind=C_CHAR), intent(in) :: name
          REAL(C_DOUBLE), intent(out) :: value
          INTEGER(C_INT) :: calceph_getconstant
          TYPE(C_PTR), VALUE, intent(in) :: eph
        END FUNCTION calceph_getconstant

        ! get the number of constants available in the ephemeris file
        FUNCTION calceph_getconstantcount(eph) BIND(C)
          USE, INTRINSIC :: ISO_C_BINDING
          IMPLICIT NONE
          TYPE(C_PTR), VALUE, intent(in) :: eph
          INTEGER(C_INT) :: calceph_getconstantcount
        END FUNCTION calceph_getconstantcount

        ! get the name and its value of the constant available at some index in the ephemeris file
        FUNCTION calceph_getconstantindex(eph, index, name, value)        &
     &                                                       BIND(C)
          USE, INTRINSIC :: ISO_C_BINDING
          IMPLICIT NONE
          TYPE(C_PTR), VALUE, intent(in) :: eph
          INTEGER(C_INT),VALUE, intent(in) :: index
          CHARACTER(len=1,kind=C_CHAR), dimension(33),intent(out)::name
          REAL(C_DOUBLE), intent(out) :: value
          INTEGER(C_INT) :: calceph_getconstantindex
        END FUNCTION calceph_getconstantindex

        ! close an ephemeris data file 
       SUBROUTINE calceph_close(eph) BIND(C)
          USE, INTRINSIC :: ISO_C_BINDING
          IMPLICIT NONE
          TYPE(C_PTR), VALUE, intent(in) :: eph
       END SUBROUTINE calceph_close


        END INTERFACE
        
        
        END MODULE calceph
